@@ -1,25 +1,26 @@
 /* dbus_service.c
- *  
+ *
  * D-BUS Service Utilities
  * Provides utilities for construction of D-BUS "Services"  
  *
- * Copyright © 2006 Red Hat, Inc.  All rights reserved.
+ * Copyright (C) 2006  Red Hat, Inc. All rights reserved.
  *
  * This copyrighted material is made available to anyone wishing to use,
  * modify, copy, or redistribute it subject to the terms and conditions of
- * the GNU General Public License v.2.  This program is distributed in the
- * hope that it will be useful, but WITHOUT ANY WARRANTY expressed or
- * implied, including the implied warranties of MERCHANTABILITY or FITNESS
- * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
- * details.  You should have received a copy of the GNU General Public
- * License along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
- * USA. Any Red Hat trademarks that are incorporated in the source code or
- * documentation are not subject to the GNU General Public License and may
- * only be used or replicated with the express permission of Red Hat, Inc.
+ * the GNU General Public License v.2, or (at your option) any later version.
+ * This program is distributed in the hope that it will be useful, but WITHOUT
+ * ANY WARRANTY expressed or implied, including the implied warranties of
+ * MERCHANTABILITY or FITNESS FOR A * PARTICULAR PURPOSE.  See the GNU General
+ * Public License for more details.  You should have received a copy of the
+ * GNU General Public License along with this program; if not, write to the
+ * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+ * 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
+ * source code or documentation are not subject to the GNU General Public
+ * License and may only be used or replicated with the express permission of
+ * Red Hat, Inc.
  *
- * Red Hat Author(s):  Jason Vas Dias
- *                     David Cantrell
+ * Red Hat Author(s): Jason Vas Dias
+ *                    David Cantrell
  */
 
 #include <stdio.h>
@@ -976,9 +977,54 @@ static void process_timeouts(DBusConnectionState * cs) {
     twalk(cs->timeouts, process_timeout);
 }
 
+static float next_timeout;
+
+static void find_timeout (const void *p, const VISIT which, const int level) {
+    DBusConnectionState *cs;
+    const void *const *tpp = p;
+    DBusConnectionTimeout *to;
+    struct timeval tv;
+    float now, then, interval, timeout;
+
+    gettimeofday(&tv, 0L);
+
+    if ((tpp != 0L) && (*tpp != 0L) && ((which == postorder) || (which == leaf))) {
+        to = (DBusConnectionTimeout *) * tpp;
+        cs = to->cs;
+
+        if (!dbus_timeout_get_enabled(to->to))
+            return;
+
+        cs = dbus_timeout_get_data(to->to);
+        then = ((float) to->tv.tv_sec) + (((float) to->tv.tv_usec) / 1000000.0);
+
+        if (then != 0.0) {
+            interval = ((float) dbus_timeout_get_interval(to->to)) / 1000.0;
+            now = ((float) tv.tv_sec) + (((float) tv.tv_usec) / 1000000.0);
+
+	    timeout = then + interval - now;
+	    if (timeout < 0)
+		timeout = 0;
+	    if (next_timeout < 0 || timeout < next_timeout)
+		next_timeout = timeout;
+        }
+    }
+}
+
+static struct timeval *find_next_timeout (DBusConnectionState * cs, struct timeval *tv) {
+    next_timeout = 1;
+    twalk(cs->timeouts, find_timeout);
+    printf ("next timeout %f\n", next_timeout);
+    if (next_timeout == -1)
+	return NULL;
+    tv->tv_sec = next_timeout;
+    tv->tv_usec = (next_timeout - tv->tv_sec) * 1000000;
+    return tv;
+}
+
 static void set_watch_fds(DBusWatch * watch, DBusConnectionState * cs) {
     uint8_t flags = dbus_watch_get_flags(watch);
-    int fd = dbus_watch_get_fd(watch);
+    int fd = dbus_watch_get_unix_fd(watch);
 
     if (cs->n <= fd)
         cs->n = fd + 1;
@@ -1011,7 +1057,7 @@ static dbus_bool_t add_watch(DBusWatch * watch, void *csp) {
     dbus_watch_set_data(watch, cs, no_free);
 
     if (cs->dh != 0L)
-        (*(cs->dh)) ("add_watch: %d", dbus_watch_get_fd (watch));
+        (*(cs->dh)) ("add_watch: %d", dbus_watch_get_unix_fd (watch));
 
     if (tsearch((const void *) watch, &(cs->watches), ptr_key_comparator) == 0L) {
         if (cs->eh != 0L)
@@ -1028,14 +1074,14 @@ static dbus_bool_t add_watch(DBusWatch * watch, void *csp) {
 static void remove_watch(DBusWatch * watch, void *csp)
 {
     DBusConnectionState *cs = csp;
-    int fd = dbus_watch_get_fd(watch);
+    int fd = dbus_watch_get_unix_fd(watch);
 
     if (tdelete((const void *) watch, &(cs->watches), ptr_key_comparator) == 0L)
         if (cs->eh != 0L)
             (*(cs->eh)) ("remove_watch: can't happen?!?: watch not found");
 
     if (cs->dh != 0L)
-        (*(cs->dh)) ("remove_watch: %d", dbus_watch_get_fd(watch));
+        (*(cs->dh)) ("remove_watch: %d", dbus_watch_get_unix_fd(watch));
 
     FD_CLR(fd, &(cs->r_fds));
     FD_CLR(fd, &(cs->w_fds));
@@ -1046,7 +1092,7 @@ static void toggle_watch(DBusWatch * watch, void *csp) {
     DBusConnectionState *cs = csp;
 
     if (cs->dh != 0L)
-        (*(cs->dh)) ("toggle_watch: %d", dbus_watch_get_fd(watch));
+        (*(cs->dh)) ("toggle_watch: %d", dbus_watch_get_unix_fd(watch));
 
     set_watch_fds(watch, cs);
 }
@@ -1068,11 +1114,11 @@ static void process_watch(const void *p, const VISIT which, const int level) {
         if (!dbus_watch_get_enabled(w))
             return;
 
-        fd = dbus_watch_get_fd(w);
+        fd = dbus_watch_get_unix_fd(w);
         flags = dbus_watch_get_flags(w);
 
         if (cs->dh != 0L)
-            (*(cs->dh)) ("handle_watch: %d", dbus_watch_get_fd(w));
+            (*(cs->dh)) ("handle_watch: %d", dbus_watch_get_unix_fd(w));
 
         if ((flags & DBUS_WATCH_READABLE) && (FD_ISSET(fd, &(cs->s_r_fds))))
             dbus_watch_handle(w, DBUS_WATCH_READABLE);
@@ -1147,10 +1193,8 @@ void dbus_svc_main_loop(DBusConnectionState * cs, void (*idle_handler) (DBusConn
         cs->s_w_fds = cs->w_fds;
         cs->s_e_fds = cs->e_fds;
 
-        timeout.tv_sec = 0;
-        timeout.tv_usec = 200000;
-
-        if ((n_fds = select(cs->n, &(cs->s_r_fds), &(cs->s_w_fds), &(cs->s_e_fds), &timeout)) < 0) {
+	if ((n_fds = select(cs->n, &(cs->s_r_fds), &(cs->s_w_fds), 
+			    &(cs->s_e_fds), find_next_timeout (cs, &timeout))) < 0) {
             if (errno != EINTR) {
                 if (cs->eh != 0L)
                     (*(cs->eh)) ("select failed: %d : %s", errno, strerror(errno));
